@file:OptIn(ExperimentalSerializationApi::class)

package crypstie3

import api.ApiError
import api.ErrorResult
import io.ktor.client.*
import io.ktor.client.call.*
import io.ktor.client.plugins.contentnegotiation.*
import io.ktor.client.request.*
import io.ktor.client.request.forms.*
import io.ktor.client.statement.*
import io.ktor.http.*
import io.ktor.serialization.kotlinx.json.*
import io.ktor.utils.io.core.*
import kotlinx.coroutines.CompletableDeferred
import kotlinx.datetime.Clock
import kotlinx.datetime.Instant
import kotlinx.serialization.ExperimentalSerializationApi
import net.sergeych.boss_serialization_mp.BossEncoder
import net.sergeych.boss_serialization_mp.decodeBoss
import net.sergeych.mp_tools.decodeBase64Compact
import net.sergeych.mp_tools.encodeToBase64Compact
import net.sergeych.mp_tools.globalLaunch
import net.sergeych.mptools.CachedExpression
import net.sergeych.unikrypto.SymmetricKeys
import tools.decodeBase64Url
import tools.encodeToBase64Url
import kotlin.time.Duration.Companion.days

class CrypstieClient(val rootUrl: String) {

    private val hc = CompletableDeferred<HttpClient>()

    init {
        globalLaunch {
            try {
                hc.complete(HttpClient() {
                    install(ContentNegotiation) {
                        json()
                    }
                })
            } catch (x: Exception) {
                hc.completeExceptionally(x)
            }
        }
    }

    private val cachedVersion = CachedExpression<String>()

    suspend fun version(): String {
        return cachedVersion.get {
            hc.await()
                .get("$rootUrl/api/version")
                .body<VersionInfo>()
                .version
        }
    }

    /**
     * Create encrypted crypstie on the server and return it as a path part of the
     * of the secret url, e.g. `_guid#haskey`. You should prepend protocol and host to your
     * web service to use it or whatever else but leave has part as a hash to not to compromise
     * your data!
     */
    suspend fun createPath(
        text: String,
        deleteAt: Instant = Clock.System.now() + 90.days,
        burnOnShow: Boolean = false,
        ownerHandle: String? = null,
    ) = createPath(text.toByteArray(), deleteAt, burnOnShow, ownerHandle = ownerHandle)

    /**
     * Create encrypted crypstie and return its path-patr of the url, e.g. `_guid#haskey`
     * just add your interface host to it
     */
    suspend fun createPath(
        data: ByteArray,
        deleteAt: Instant = Clock.System.now() + 90.days,
        burnOnShow: Boolean = false,
        profileIds: List<String>? = null,
        useWide: Boolean = false,
        syntaxHighlighting: String? = null,
        editable: Boolean = false,
        ownerHandle: String? = null,
    ): String {
        // Generate random key and encrypt data
        val k = SymmetricKeys.random()
        val encrypted = k.etaEncrypt(data)
        val keyAsString = k.keyBytes.encodeToBase64Url()

        // post crypstie and return URL from the service
        val response = hc.await().submitFormWithBinaryData(
            url = "$rootUrl/api/crypstie",
            formData = formData {
                append(
                    "data",
                    BossEncoder.encode(
                        ApiCrypstie(
                            "",
                            encrypted, deleteAt, burnOnShow, useWide,
                            syntaxHighlighting, editable, ownerHandle
                        )
                    ),
                    Headers.build {
                        append(HttpHeaders.ContentType, "application/octet-stream")
                        append(HttpHeaders.ContentDisposition, "filename=\"dummy.bin\"")
                    }
                )
            }
        )
        val guid = checkResponse(response).body<String>()
        // now complete it with a key:
        return "_$guid#$keyAsString"
    }

    private suspend fun checkResponse(response: HttpResponse): HttpResponse {
        if (!response.status.isSuccess()) {
            if (response.status.value == 404)
                ApiError.INTERNAL_ERROR.raise("404: not found")
            try {
                response.body()
            } catch (x: Exception) {
                ErrorResult(ApiError.UNKNOWN_ERROR, x.toString())
            }.raise()
        }
        return response
    }

    /**
     * donwload and decrypt crypstie by its url path part with hashtah
     * @param pathPart path part of the url e.g. `_<id>#<encoded_key>`.
     * @throws ApiError if can't download/decrypt.
     */
    suspend fun decryptPath(pathPart: String, profileIds: List<String>? = null): DecryptedCrypstie {
        if (pathPart[0] != '_') ApiError.BAD_PARAMETER.raise("pathPart should start with #")
        val parts = pathPart.split('#')

        fun invalidFormat(text: String? = null): Nothing {
            ApiError.BAD_PARAMETER.raise(
                "invalid pathPart format${text?.let { ": $it" } ?: ""} "
            )
        }

        if (parts.size != 2 || parts[0].isBlank() || parts[1].isBlank())
            invalidFormat()
        val key = try {
            val keyBytes = parts[1].decodeBase64Url()
            if (keyBytes.size != 32) invalidFormat("wrong hashkey size")
            SymmetricKeys.create(keyBytes)
        } catch (x: Exception) {
            invalidFormat("failed to decode hashkey: $x")
        }

        val query = if (profileIds.isNullOrEmpty()) ""
        else "?pids=${profileIds.joinToString(",")}"
        val packed = checkResponse(hc.await().get("$rootUrl/api/crypstie/${parts[0].substring(1)}$query"))
            .body<ByteArray>()
        val ac = packed.decodeBoss<ApiCrypstie>()
        try {
            return DecryptedCrypstie(key, ac)
        } catch (x: Exception) {
            invalidFormat("failed to decrypt")
        }
    }

}