package tools

import kotlinx.serialization.KSerializer
import kotlinx.serialization.json.Json
import kotlinx.serialization.serializer
import kotlin.reflect.KProperty

interface BackingStorge {
    fun getItem(key: String): String?
    fun setItem(key: String, value: String)
    fun clearItem(key: String)
}

class TempBackingStorage : BackingStorge {

    private val m = mutableMapOf<String,String>()

    override fun getItem(key: String): String? = m[key]

    override fun setItem(key: String, value: String) {
        m[key]  = value
    }

    override fun clearItem(key: String) {
        m.remove(key)
    }
}

class PrefixedStorage(private val bs: BackingStorge, val prefix: String): BackingStorge {

    override fun getItem(key: String): String? {
        return bs.getItem(prefix + key)
    }

    override fun setItem(key: String, value: String) {
        bs.setItem(prefix+key, value)
    }

    override fun clearItem(key: String) {
        bs.clearItem(prefix+key)
    }

}

private val format = Json { prettyPrint = false }

inline fun <reified T>storage(backingStorge: BackingStorge,overrideName: String?=null): SimpleStorageDelegate<T> {
    return SimpleStorageDelegate(backingStorge, overrideName, serializer())
}

class SimpleStorageDelegate<T>(
    val storage: BackingStorge,
    val overrideName: String? = null,
    val sr: KSerializer<T>,
) {
    private var cacheIsSet = false
    private var cachedValue: T? = null

    operator fun getValue(thisRef: Any?, property: KProperty<*>): T? =
        if (cacheIsSet)
            cachedValue
        else
            storage.getItem(overrideName ?: property.name)?.let { format.decodeFromString(sr, it) }

    operator fun setValue(thisRef: Any?, property: KProperty<*>, value: T?) {
        val name = overrideName ?: property.name
        if (value == null) storage.clearItem(name)
        else storage.setItem(name, format.encodeToString(sr, value))
        cachedValue = value
        cacheIsSet = true
    }
}
